'''
Raet package metadata
'''

__version_info__ = (0, 3, 4)
__version__ = '{0}.{1}.{2}'.format(*__version_info__)
__author__ = "Samuel M. Smith"
__license__ = "Apache2"

__all__ = ['__version_info__', '__version__', '__author__', '__license__']
